/*
   xvgen - Signal generator on X11(XView) or command line
   written by Akira NISHIMURA 1998/1/7
*/

#include <xview/xview.h>
#include <xview/frame.h>
#include <xview/panel.h>
#include <xview/notify.h>
#include <xview/openmenu.h>
#include <xview/notice.h>   /* about xvgen */
#include <xview/cms.h>
#include <xview/icon.h>     /* for icon */
#include <sys/time.h>
#include <stdio.h>
#include <unistd.h>   /* for getopt() */
#include <math.h>
#include <stdlib.h>
#include <time.h>     /* for time() */
#include <sys/types.h>
#include <unistd.h>
#ifdef _FREEBSD
#include <machine/soundcard.h>
#else
#include <linux/soundcard.h>
#endif
#include "xvsnd.h"

#define TITLE           "xvgenerator ver. 0.2"
#define ABOUT_NW_XVGEN  "Signal generator on command line"
#define ABOUT_XVGEN     "Signal generator on X11(XView) or command line"
#define COPY_RIGHT      "Copyright (C) 1998, Akira Nishimura"
#define INIT_MESSG      "Xvgen is now ready to generate"
#define AUTHOR_ADDRESS  "akira@rsch.tuis.ac.jp"

#define MAX_SF		(44100)  /* max sampling freq. in Hz */
#define MIN_SF		(4000)   /* min sampling freq. in Hz */
#define MAX_AMP		(0)      /* max amplitude in dB */
#define MIN_AMP  	{-30,-60}/* min amplitude in dB for 8bits and 16bits */
#define MIN_DB          {-42,-90}/* min amplitude */
#define WAIT_MSEC	(30)     /* waiting time in timer in msec */

/* sampling frequency sereices */
#define SF_STR		"11025", "22050", "44100"
#define SF_NUM		(3)
#define BITS8_16        "8", "16"
#define BITS8           "8"
#define DEFAULT_SF	(11025)  /* default sampling freq. in Hz */
#define DEFAULT_SB	(16)     /* default sampling resolution in bits */
#define DEFAULT_FREQ	(1000)	 /* default freq. of pure tone in Hz */
#define DEFAULT_AMP	(-10)	 /* default amplitude in dB */
#define DEFAULT_CH	(1)      /* default number of channel */
#define DEFAULT_MODE    (0)      /* default execution mode */
#define DEFAULT_TYPE    (0)      /* default signal type 0:pure tone, 1:noise */

#define LEVEL_CAUTION   "Caution!! Amplitude level is not reliable!!"

#ifdef _USE_XVIEW

Frame       frame;
Panel       panel;
Panel_item  message, sf_item, freq_item, srb_item, amp_item;
int         min_amp[] = MIN_AMP;
struct itimerval timer;


#include    "xvgen.xbm"          /* icon file */

#else

#define     Notify_value     int

#endif

int         dsp;             /* dsp identifier */
float       amp;             /* amplitude in dB */
short      *abuf;            /* audio buffer for 16-bit */
unsigned char *abuf8;         /* audio buffer for 8-bit */
float       phase;           /* phase of a pure tone in radian */
int         freq;            /* frequency in Hz */
int         period;          /* length of audio buffer */
int         channel;         /* 0:center, 1:left;  2; right */
int         stype;           /* type of signal, 0:pure_tone, 1:white_noise */
float       min_db[] = MIN_DB;
char       *sf_str[] = {SF_STR};
char        dsp_name[12];        /* device name */
snd_head    wav;
char        messg[80] = INIT_MESSG;

int         out_flag;         /* flag for output sounds, 0: not out 1:out */
FILE       *snd_out;

#ifdef _Debug
FILE *debugtext;
#endif

Notify_value play_sound()
{

  switch(stype){
  case 0:
#ifdef _Debug
    fprintf(debugtext, "f %d s %d p %d a %f p %f\n",
	    freq, wav.sf, period, amp, phase);
#endif
    mkptone(abuf, freq, wav.sf, period, amp, &phase);
    break;
  case 1:
    mkwnoise(abuf, period, amp);
    break;
  }
  if(wav.srb == 8){
    b16to8(abuf, abuf8, period);
    write(dsp, abuf8, period);
  }
  else{
    write(dsp, abuf, period*2);
  }

  if(out_flag){
    fwrite(abuf, sizeof(short), period, snd_out);
  }

/*  ioctl(dsp, SNDCTL_DSP_SYNC, NULL); */

  return NOTIFY_DONE;
}

#ifdef _USE_XVIEW

void timer_sound(Panel_item item)
{
  int value;
  snd_head set_wav;

  value = xv_get(item, PANEL_VALUE);

#ifdef _Debug_Status
  fprintf(stderr, "Gen Status: %d selected\n", value);
#endif

  if(value == 0){
    if((dsp = dsp_wopen(dsp_name)) < 0){
      exit(-1);
    }
    set_wav = set_dsp_param(dsp, wav);
    if(set_wav.sf != wav.sf){
      sprintf(messg, "Real Sampling Freq. is %d Hz", set_wav.sf);
      wav = set_wav;
      xv_set(message, PANEL_LABEL_STRING, messg, NULL);
    }

    timer.it_value.tv_usec = timer.it_interval.tv_usec =
      (int)floor((double)period/wav.sf*1000000)-WAIT_MSEC*1000;

#ifdef _Debug_Status
    fprintf(stderr, "Wait: %d\n", timer.it_interval.tv_usec);
#endif
    notify_set_itimer_func(frame, play_sound, ITIMER_REAL, &timer, NULL);
  }
  else{
    notify_set_itimer_func(frame, NOTIFY_FUNC_NULL, ITIMER_REAL, NULL, NULL);
    ioctl(dsp, SOUND_PCM_RESET, 0);
    close(dsp); 
    dsp = -1;
  }
}

int set_freq(Panel_item item)
{
  int oldfreq;

  oldfreq = freq;
  freq = xv_get(item, PANEL_VALUE);

  if(freq > wav.sf/2 || freq < 0){
    sprintf(messg, "Improper Freq. %d[Hz], set to %d", freq, oldfreq);
    freq = oldfreq;
    xv_set(message, PANEL_LABEL_STRING, messg, NULL);
    xv_set(freq_item, PANEL_VALUE, oldfreq, NULL);
    fprintf(stderr, "%s\n", messg);
    return XV_ERROR;
  }
  else{
    sprintf(messg, "Freq. %d Hz", freq);
    xv_set(message, PANEL_LABEL_STRING, messg, NULL);
  }

#ifdef _Debug_Status
  fprintf(stderr, "freq: %d\n", freq);
#endif

  return XV_OK;
}

int set_ch(Panel_item item)
{
  int oldch;

  channel = xv_get(item, PANEL_VALUE);
#ifdef _Debug_Status
  fprintf(stderr, "ch: %d\n", channel);
#endif

  oldch = wav.ch;
  if(channel > 0){
    wav.ch = 2;
  }
  else{
    wav.ch = 1;
  }
  if(oldch != wav.ch){
    set_dsp_param(dsp, wav);
    free(abuf);
    free(abuf8);
    abuf = malloc(period*2*sizeof(short));
    abuf8 = malloc(period*2*sizeof(unsigned char));
  }
  return XV_OK;
}

int set_sf(Panel_item item, int value)
{
  int oldsf, i;
  snd_head set_wav;

  oldsf = wav.sf;
  wav.sf = atoi(sf_str[value]);
  for(i=0; i<SF_NUM; i++){
    if(oldsf == atoi(sf_str[i])) break;
  }

  if(oldsf != wav.sf){
    if(wav.sf < freq*2){
      sprintf(messg, "Improper Sampling Freq. %d[Hz]. Set to %d[Hz]",
	      wav.sf, oldsf);
      xv_set(message, PANEL_LABEL_STRING, messg, NULL);
      xv_set(sf_item, PANEL_VALUE, i, NULL);
      fprintf(stderr, "%s\n", messg);
      wav.sf = oldsf;
      return XV_ERROR;
    }
    else{
      if(dsp > 1){
	ioctl(dsp, SNDCTL_DSP_SYNC, NULL);
	set_wav = set_dsp_param(dsp, wav);
      }
      else{
	set_wav = wav;
      }
      if(set_wav.sf != wav.sf){
	sprintf(messg, "Real Sampling Freq. is %d Hz", set_wav.sf);
	wav = set_wav;
      }
      else{
	sprintf(messg, "Sampling Freq. %d Hz", wav.sf);
      }
      xv_set(message, PANEL_LABEL_STRING, messg, NULL);
      xv_set(freq_item, PANEL_MAX_VALUE, wav.sf/2, NULL);

      period = period*wav.sf/oldsf;
      free(abuf);
      free(abuf8);
      abuf = malloc(period*sizeof(short));
      abuf8 = malloc(period*sizeof(unsigned char));

#ifdef _Debug_Status
      fprintf(stderr, "SF: %d (Hz)\n", wav.sf);
#endif
    }
  }
  return XV_OK;
}

int set_srb(Panel_item item, int value)
{
  int oldsrb;

  oldsrb = wav.srb;
  wav.srb = (value+1)*8;

  if(amp < min_db[wav.srb == 8 ? 0 : 1]){
    sprintf(messg, "Amplitude is too small for %d-bit mode",
	    wav.srb);
    xv_set(message, PANEL_LABEL_STRING, messg, NULL);
    wav.srb = oldsrb;
  }

  if(oldsrb != wav.srb){
    if(dsp > 1){
      ioctl(dsp, SNDCTL_DSP_SYNC, NULL);
      set_dsp_param(dsp, wav);
    }
  
    sprintf(messg, "Sampling Resolution %d-bit", wav.srb);

    xv_set(amp_item, PANEL_MIN_VALUE, min_amp[value], NULL);
    xv_set(amp_item, PANEL_TICKS,(MAX_AMP-min_amp[value])/10+1, NULL);
    xv_set(message, PANEL_LABEL_STRING, messg, NULL);
  }

  return XV_OK;
}

int set_amp(Panel_item item)
{
  int amp_int;
  float oldamp;

  oldamp = amp;
  amp_int = xv_get(item, PANEL_VALUE);
  amp = (float)amp_int;

  if(amp > MAX_AMP || amp < min_db[wav.srb == 8 ? 0 : 1]){
    sprintf(messg, "Improper amplitude %g, set to %g dB\n", amp, oldamp);
    xv_set(message, PANEL_LABEL_STRING, messg, NULL);
    amp = oldamp;
  }

  if(stype == 1 && amp > -5.0){
    xv_set(message, PANEL_LABEL_STRING, LEVEL_CAUTION, NULL);
    fprintf(stderr, "%s\n", LEVEL_CAUTION);
  }
  else{
    xv_set(message, PANEL_LABEL_STRING, NULL, NULL);
  }

#ifdef _Debug_Status
  fprintf(stderr, "Amp: %f (dB)\n", amp);
#endif
  
  return XV_OK;
}

int set_type(Panel_item item)
{
  stype = xv_get(item, PANEL_VALUE);
#ifdef _Debug_Status
  fprintf(stderr, "Type: %d (dB)\n", stype);
#endif
  return XV_OK;
}

void about_xvgen()
{
  xv_create(panel, NOTICE,
	    NOTICE_MESSAGE_STRINGS,
	      TITLE,
	      ABOUT_XVGEN,
	      COPY_RIGHT,
	      AUTHOR_ADDRESS,
	    NULL,
	    XV_SHOW, TRUE,
	    NULL);
}

void quit()
{
  xv_destroy_safe(frame);
}

#endif /* end of #define _USE_XVIEW */

int init_abuf()
{
  int abuf_size;
/*
  int frag = 8192; 

  ioctl(dsp, SNDCTL_DSP_SETFRAGMENT, &frag);
*/
  ioctl(dsp, SNDCTL_DSP_GETBLKSIZE, &abuf_size);
  abuf_size = 16384;
  abuf = malloc(abuf_size);
  abuf8 = malloc(abuf_size/2);
  return abuf_size/sizeof(short);
}

void for_help()
{
#ifdef _USE_XVIEW
  printf("xvgen : %s.\n", ABOUT_XVGEN);
#else
  printf("xvgen : %s.\n", ABOUT_NW_XVGEN);
#endif
  printf("%s  %s\n", COPY_RIGHT, AUTHOR_ADDRESS);
  printf("  available options :\n");
#ifndef _USE_XVIEW
  printf("  -nw               : no window mode (default : window mode)\n");
#endif
  printf("  -s [freq.]        : sampling freq. [Hz] (default : %d)\n",
	DEFAULT_SF);
  printf("  -f [freq.]        : freq. of a pure tone [Hz] (default : %d)\n",
	DEFAULT_FREQ);
  printf("  -b [8 or 16]      : sampling resolution [bits] (default : %d)\n",
	DEFAULT_SB);
  printf("  -a [amp.]         : amplitude [dB] (default : %d)\n",
	DEFAULT_AMP);
  printf("  -t [signal type]  : 0: pure tone, 1: white noise (default : %d)\n",
	DEFAULT_TYPE);
  printf("  -o                : output binary-short stream to STDOUT (default : not out)\n");
  printf("  -v                : show version of xvgen\n");
}

int main(int argc, char *argv[])
{
#ifdef _USE_XVIEW
  Panel_item quit_item;
  Rect *rect;
  Icon            icon;
  Pixmap          icon_pm;
  Server_image    icon_image;
  Display        *dpy;
  Window          root_win;
  int             srb_value;
#endif
  int no_window, i, srb_8_16, srb_probe;
  snd_head set_wav;
  unsigned int seed;
  char c;

  seed = ((unsigned int)time(NULL) | (unsigned int)getpid());
  srand(seed);
  srand48(seed);

#ifdef _USE_XVIEW
  xv_init(XV_INIT_ARGC_PTR_ARGV, &argc, argv, NULL);
#endif
  
  strcpy(dsp_name, "/dev/dsp");
  if(!strcmp(argv[0], "xvgen2")){
    strcpy(dsp_name, "/dev/dsp1");
  }

  wav.sf = DEFAULT_SF;
  wav.srb = DEFAULT_SB;
  wav.ch = DEFAULT_CH;
  freq = DEFAULT_FREQ;
  amp =  DEFAULT_AMP;
  channel = DEFAULT_CH;
  stype = DEFAULT_TYPE;
  no_window = DEFAULT_MODE;
  out_flag = 0;

  while ((c = getopt (argc, argv, "n:s:f:b:a:t:ovh")) != EOF){
    switch (c){
    case 'n':
      if(!strcmp(optarg, "w")) no_window = 1;
      else                     no_window = 0;
      break;
    case 's':
      wav.sf = atoi(optarg);
      if(wav.sf < MIN_SF || wav.sf > MAX_SF){
	fprintf(stderr, "Sampling freq. must be %d -- %d\n",
		MIN_SF, MAX_SF);
	exit(-1);
      }
      break;
    case 'f':
      freq = atoi(optarg);
      if(freq < 0 || freq > MAX_SF){
	fprintf(stderr, "Frequency must be positive and less than %d\n",
		MAX_SF);
	exit(-1);
      }
      break;
    case 'b':
      wav.srb = atoi(optarg);
      if(wav.srb != 8 && wav.srb != 16){
	fprintf(stderr, "Sampling resolution must be 8 or 16\n");
	wav.srb = 16;
      }
      break;
    case 'a':
      amp = atof(optarg);
      break;
    case 't':
      stype = atoi(optarg);
      if(stype < 0 && stype > 1) stype = DEFAULT_TYPE;
      break;
    case 'o':
      out_flag = 1;
      snd_out = stdout;
      break;
    case 'v':
#ifdef _USE_XVIEW
      fprintf(stderr, "%s\n", TITLE);
#else
      fprintf(stderr, "%s (command line only)\n", TITLE);
#endif
      exit(1);
      break;
    case 'h':
    default:
      for_help();
      exit(1);
      break;
    }
  }

  for(i=0; i<SF_NUM; i++){
    if(wav.sf == atoi(sf_str[i])) break;
  }
  if(i == SF_NUM) i = DEFAULT_SF;
  if(amp > -5.0 && stype == 1){
    fprintf(stderr, "%s\n", LEVEL_CAUTION);
  }
  if(freq > wav.sf/2){
    fprintf(stderr, "Frequency must be less than %d", wav.sf/2);
    exit(-1);
  }

#ifndef _USE_XVIEW
  no_window = 1;
#endif

  if((dsp = dsp_wopen(dsp_name)) < 0){
    exit(-1);
  }
  srb_probe = 16;
  if(ioctl(dsp, SOUND_PCM_WRITE_BITS, &srb_probe) == -1){
    fprintf(stderr, "Error. Can't change sampling resolution to %d bits\n",
	    wav.srb);
    exit(-1);
  }
  if(srb_probe != 16){  /* sound card has only 8-bit mode */
    srb_8_16 = 0;
  }
  else{                 /* sound card has both 8,16-bit mode */
    srb_8_16 = 1;
  }
    
  if(amp > MAX_AMP || amp < min_db[wav.srb == 8 ? 0 : 1]){
    fprintf(stderr, "Amplitude must be %g -- %d dB\n",
	    min_db[wav.srb == 8 ? 0 : 1], MAX_AMP);
    exit(-1);
  }

  period = init_abuf();

#ifdef _Debug_Status
  fprintf(stderr, "abuf_size : %d\n", period);
#endif

#ifdef _Debug  
  debugtext = fopen("debug.txt", "w");
#endif

  set_wav = set_dsp_param(dsp, wav);
  if(set_wav.sf != wav.sf){
    sprintf(messg, "Real Sampling Freq. is %d Hz", set_wav.sf);
  }
  wav = set_wav;

  if(no_window){
    switch(stype){
    case 0:
      sprintf(messg, "Pure tone, Freq. = %d Hz,", freq);
      break;
    case 1:
      sprintf(messg, "White noise,");
      break;
    }
    sprintf(messg, "%s  Amp. = %g dB,  Sampl. freq. = %d Hz, %d-bit",
	    messg, amp, wav.sf, wav.srb);
    fprintf(stderr, "%s\n  ctrl+c to exit\n", messg);
    while(1){
      play_sound();
    }
  }
  else{
#ifdef _USE_XVIEW
    close(dsp);
    dsp = -1;
    frame = (Frame)xv_create(NULL, FRAME,
			   FRAME_LABEL, TITLE,
			   XV_WIDTH,    400,
			   XV_HEIGHT,   300,
			   NULL);

    panel = (Panel)xv_create(frame, PANEL,
			   PANEL_LAYOUT,
			   PANEL_VERTICAL,
			   NULL);

    sf_item = (Panel_item)xv_create(panel, PANEL_CHOICE,
            PANEL_DISPLAY_LEVEL, PANEL_CURRENT,		      
	    PANEL_LAYOUT,       PANEL_HORIZONTAL,
	    PANEL_LABEL_STRING, "Sampling Freq. (Hz) : ",
            PANEL_CHOICE_STRINGS, SF_STR, NULL,
	    PANEL_NOTIFY_PROC,   set_sf,
	    PANEL_VALUE,         i,
	    NULL);

    srb_value = (wav.srb-8)/8;
    if(srb_8_16){
      srb_item = (Panel_item)xv_create(panel, PANEL_CHOICE,
            PANEL_DISPLAY_LEVEL, PANEL_CURRENT,		      
	    PANEL_LAYOUT,       PANEL_HORIZONTAL,
	    PANEL_LABEL_STRING, "Sampling Resolution (bits) : ",
            PANEL_CHOICE_STRINGS, BITS8_16, NULL,
	    PANEL_NOTIFY_PROC,   set_srb,
	    PANEL_VALUE,         srb_value,
	    NULL);
    }
    else{
      srb_item = (Panel_item)xv_create(panel, PANEL_CHOICE,
            PANEL_DISPLAY_LEVEL, PANEL_CURRENT,		      
	    PANEL_LAYOUT,       PANEL_HORIZONTAL,
	    PANEL_LABEL_STRING, "Sampling Resolution (bits) : ",
            PANEL_CHOICE_STRINGS, BITS8, NULL,
	    PANEL_NOTIFY_PROC,   set_srb,
	    PANEL_VALUE,         0,
	    NULL);
    }

    freq_item = xv_create(panel, PANEL_NUMERIC_TEXT,
	    PANEL_LAYOUT,       PANEL_HORIZONTAL,
	    PANEL_LABEL_STRING, "Frequency (Hz) : ",
	    PANEL_VALUE,        freq,
	    PANEL_MIN_VALUE,     1,
	    PANEL_MAX_VALUE,     wav.sf/2,
	    PANEL_NOTIFY_PROC,  set_freq,
	    NULL);

    amp_item = xv_create(panel, PANEL_SLIDER,
	    PANEL_LAYOUT,       PANEL_HORIZONTAL,
	    PANEL_LABEL_STRING, "Amplitude (dB) : ",
	    PANEL_VALUE,        (short)amp,
	    PANEL_MIN_VALUE,    min_amp[srb_value],
	    PANEL_MAX_VALUE,    MAX_AMP,
	    PANEL_SLIDER_WIDTH, -min_amp[1]*2,
	    PANEL_TICKS,        (MAX_AMP-min_amp[srb_value])/10+1,
	    PANEL_NOTIFY_PROC,  set_amp,
	    NULL);

    xv_create(panel, PANEL_CHOICE,
            PANEL_DISPLAY_LEVEL, PANEL_CURRENT,		      
	    PANEL_LAYOUT,       PANEL_HORIZONTAL,
	    PANEL_LABEL_STRING, "Signal Type : ",
            PANEL_CHOICE_STRINGS, "Pure Tone", "White Noise", NULL,
	    PANEL_NOTIFY_PROC,   set_type,
	    PANEL_VALUE,         stype,
	    NULL);

    xv_create(panel, PANEL_CHOICE,
	    PANEL_LABEL_STRING, "Generation",
	    PANEL_CHOICE_STRINGS, "START", "STOP", NULL,
	    PANEL_NOTIFY_PROC,   timer_sound,
	    PANEL_VALUE,         1,
	    NULL);

    quit_item = xv_create (panel, PANEL_BUTTON,
            PANEL_LABEL_STRING,         "Quit",
            PANEL_NOTIFY_PROC,          quit,
            NULL);

    rect = (Rect *)xv_get(quit_item, XV_RECT);
    
    xv_create(panel, PANEL_BUTTON,
	    PANEL_LABEL_STRING,     "About Xvgen",
	    PANEL_NOTIFY_PROC,      about_xvgen,
	    XV_X,       rect_right(rect)+30,
	    XV_Y,       rect_bottom(rect)-17,
	    NULL);

    message = xv_create(panel, PANEL_MESSAGE,
		      PANEL_LABEL_STRING, messg,
		      NULL);

/* for icon image */
    dpy = (Display *) xv_get(frame, XV_DISPLAY);
    root_win = RootWindow (dpy, DefaultScreen(dpy));
    icon_pm = XCreatePixmapFromBitmapData(dpy,
		root_win, xvgen_bits, xvgen_width, xvgen_height,
                1, 0, 1);
    icon_image = (Server_image) xv_create(NULL,SERVER_IMAGE,
                        XV_WIDTH,   xvgen_width,
                        XV_HEIGHT,  xvgen_height,
                        SERVER_IMAGE_PIXMAP, icon_pm, NULL);
    icon = (Icon) xv_create(frame, ICON,
                        XV_WIDTH,   xvgen_width,
                        XV_HEIGHT,  xvgen_height,
                        ICON_IMAGE, icon_image,
                        NULL);
    xv_set(frame,
	   FRAME_ICON, icon,
	   NULL);
/* end of icon image */

    window_fit(panel);
    window_fit(frame);
    xv_main_loop(frame);
#endif /* end of #ifdef _USE_XVIEW */
  }
  close(dsp);
  exit(0);
}
